@@ -16,30 +16,30 @@
 
 """Trac Environment model and related APIs."""
 
-from __future__ import with_statement
-
+import hashlib
 import os.path
 import setuptools
+import shutil
 import sys
 from urlparse import urlsplit
 
-from trac import db_default
+from trac import db_default, log
 from trac.admin import AdminCommandError, IAdminCommandProvider
 from trac.cache import CacheManager, cached
-from trac.config import BoolOption, ConfigSection, Configuration, Option, \
-                        PathOption
+from trac.config import BoolOption, ChoiceOption, ConfigSection, \
+                        Configuration, Option, PathOption
 from trac.core import Component, ComponentManager, implements, Interface, \
-                      ExtensionPoint, TracError
+                      ExtensionPoint, TracBaseError, TracError
 from trac.db.api import (DatabaseManager, QueryContextManager,
                          TransactionContextManager, with_transaction)
-from trac.util import copytree, create_file, get_pkginfo, lazy, makedirs, \
-                      read_file
-from trac.util.compat import sha1
+from trac.loader import load_components
+from trac.log import logger_handler_factory
+from trac.util import arity, as_bool, copytree, create_file, get_pkginfo, \
+                      lazy, makedirs, read_file
 from trac.util.concurrency import threading
 from trac.util.text import exception_to_unicode, path_to_unicode, printerr, \
                            printout
 from trac.util.translation import _, N_
-from trac.versioncontrol import RepositoryManager
 from trac.web.href import Href
 
 __all__ = ['Environment', 'IEnvironmentSetupParticipant', 'open_environment']
@@ -74,15 +74,19 @@ class IEnvironmentSetupParticipant(Interface):
     def environment_created():
         """Called when a new Trac environment is created."""
 
-    def environment_needs_upgrade(db):
+    def environment_needs_upgrade(db=None):
         """Called when Trac checks whether the environment needs to be
         upgraded.
 
         Should return `True` if this participant needs an upgrade to
         be performed, `False` otherwise.
+
+        :since 1.1.2: the `db` parameter is deprecated and will be removed
+                      in Trac 1.3.1. A database connection should instead be
+                      obtained using a context manager.
         """
 
-    def upgrade_environment(db):
+    def upgrade_environment(db=None):
         """Actually perform an environment upgrade.
 
         Implementations of this method don't need to commit any
@@ -93,10 +97,14 @@ class IEnvironmentSetupParticipant(Interface):
         However, if the `upgrade_environment` consists of small,
         restartable, steps of upgrade, it can decide to commit on its
         own after each successful step.
+
+        :since 1.1.2: the `db` parameter is deprecated and will be removed
+                      in Trac 1.3.1. A database connection should instead be
+                      obtained using a context manager.
         """
 
 
-class BackupError(RuntimeError):
+class BackupError(TracBaseError, RuntimeError):
     """Exception raised during an upgrade when the DB backup fails."""
 
 
@@ -140,13 +148,13 @@ class Environment(Component, ComponentManager):
         {{{
         [components]
         trac.ticket.report.ReportModule = disabled
-        webadmin.* = enabled
+        acct_mgr.* = enabled
         }}}
 
         The first option tells Trac to disable the
         [wiki:TracReports report module].
         The second option instructs Trac to enable all components in
-        the `webadmin` package. Note that the trailing wildcard is
+        the `acct_mgr` package. Note that the trailing wildcard is
         required for module/package matching.
 
         To view the list of active components, go to the ''Plugins''
@@ -162,8 +170,7 @@ class Environment(Component, ComponentManager):
         Plugins in that directory are loaded in addition to those in
         the directory of the environment `plugins`, with this one
         taking precedence.
-
-        (''since 0.11'')""")
+        """)
 
     base_url = Option('trac', 'base_url', '',
         """Reference URL for the Trac deployment.
@@ -174,7 +181,7 @@ class Environment(Component, ComponentManager):
         resources in notification e-mails.""")
 
     base_url_for_redirect = BoolOption('trac', 'use_base_url_for_redirect',
-            False,
+                                        False,
         """Optionally use `[trac] base_url` for redirects.
 
         In some configurations, usually involving running Trac behind
@@ -183,15 +190,16 @@ class Environment(Component, ComponentManager):
         force Trac to use the `base_url` setting also for
         redirects. This introduces the obvious limitation that this
         environment will only be usable when accessible from that URL,
-        as redirects are frequently used. ''(since 0.10.5)''""")
+        as redirects are frequently used.
+        """)
 
     secure_cookies = BoolOption('trac', 'secure_cookies', False,
         """Restrict cookies to HTTPS connections.
 
         When true, set the `secure` flag on all cookies so that they
         are only sent to the server on HTTPS connections. Use this if
-        your Trac instance is only accessible through HTTPS. (''since
-        0.11.2'')""")
+        your Trac instance is only accessible through HTTPS.
+        """)
 
     project_name = Option('project', 'name', 'My Project',
         """Name of the project.""")
@@ -213,7 +221,8 @@ class Environment(Component, ComponentManager):
 
         This can be an absolute or relative URL, or '.' to reference
         this Trac instance. An empty value will disable the reporting
-        buttons.  (''since 0.11.3'')""")
+        buttons.
+        """)
 
     project_footer = Option('project', 'footer',
                             N_('Visit the Trac open source project at<br />'
@@ -224,7 +233,8 @@ class Environment(Component, ComponentManager):
     project_icon = Option('project', 'icon', 'common/trac.ico',
         """URL of the icon of the project.""")
 
-    log_type = Option('logging', 'log_type', 'none',
+    log_type = ChoiceOption('logging', 'log_type',
+                            log.LOG_TYPES + log.LOG_TYPE_ALIASES,
         """Logging facility to use.
 
         Should be one of (`none`, `file`, `stderr`, `syslog`, `winlog`).""")
@@ -234,33 +244,35 @@ class Environment(Component, ComponentManager):
         log-file.  Relative paths are resolved relative to the `log`
         directory of the environment.""")
 
-    log_level = Option('logging', 'log_level', 'DEBUG',
+    log_level = ChoiceOption('logging', 'log_level',
+                             tuple(reversed(log.LOG_LEVELS)) +
+                             log.LOG_LEVEL_ALIASES,
         """Level of verbosity in log.
 
-        Should be one of (`CRITICAL`, `ERROR`, `WARN`, `INFO`, `DEBUG`).""")
+        Should be one of (`CRITICAL`, `ERROR`, `WARNING`, `INFO`, `DEBUG`).
+        """)
 
     log_format = Option('logging', 'log_format', None,
         """Custom logging format.
 
         If nothing is set, the following will be used:
 
-        Trac[$(module)s] $(levelname)s: $(message)s
+        `Trac[$(module)s] $(levelname)s: $(message)s`
 
-        In addition to regular key names supported by the Python
-        logger library (see
-        http://docs.python.org/library/logging.html), one could use:
+        In addition to regular key names supported by the
+        [http://docs.python.org/library/logging.html Python logger library]
+        one could use:
 
-        - $(path)s     the path for the current environment
-        - $(basename)s the last path component of the current environment
-        - $(project)s  the project name
+        - `$(path)s`     the path for the current environment
+        - `$(basename)s` the last path component of the current environment
+        - `$(project)s`  the project name
 
         Note the usage of `$(...)s` instead of `%(...)s` as the latter form
-        would be interpreted by the ConfigParser itself.
+        would be interpreted by the !ConfigParser itself.
 
         Example:
         `($(thread)d) Trac[$(basename)s:$(module)s] $(levelname)s: $(message)s`
-
-        ''(since 0.10.5)''""")
+        """)
 
     def __init__(self, path, create=False, options=[]):
         """Initialize the Trac environment.
@@ -275,17 +287,32 @@ class Environment(Component, ComponentManager):
         ComponentManager.__init__(self)
 
         self.path = path
+        self.log = None
+        self.config = None
+        self._log_handler = None
+        # System info should be provided through ISystemInfoProvider rather
+        # than appending to systeminfo, which may be a private in a future
+        # release.
         self.systeminfo = []
 
         if create:
             self.create(options)
+            for setup_participant in self.setup_participants:
+                setup_participant.environment_created()
         else:
             self.verify()
             self.setup_config()
 
-        if create:
-            for setup_participant in self.setup_participants:
-                setup_participant.environment_created()
+    def __repr__(self):
+        return '<%s %r>' % (self.__class__.__name__, self.path)
+
+    @lazy
+    def name(self):
+        """The environment name.
+
+        :since: 1.2
+        """
+        return os.path.basename(self.path)
 
     @property
     def env(self):
@@ -296,26 +323,26 @@ class Environment(Component, ComponentManager):
         return self
 
     def get_systeminfo(self):
-        """Return a list of `(name, version)` tuples describing the
-        name and version information of external packages used by Trac
-        and plugins.
+        """Return a list of `(name, version)` tuples describing the name
+        and version information of external packages used by Trac and plugins.
         """
         info = self.systeminfo[:]
         for provider in self.system_info_providers:
             info.extend(provider.get_system_info() or [])
-        info.sort(key=lambda (name, version): (name != 'Trac', name.lower()))
-        return info
+        return sorted(set(info),
+                      key=lambda (name, ver): (name != 'Trac', name.lower()))
 
     # ISystemInfoProvider methods
 
     def get_system_info(self):
-        from trac import core, __version__ as VERSION
-        yield 'Trac', get_pkginfo(core).get('version', VERSION)
+        yield 'Trac', self.trac_version
         yield 'Python', sys.version
         yield 'setuptools', setuptools.__version__
         from trac.util.datefmt import pytz
         if pytz is not None:
             yield 'pytz', pytz.__version__
+        if hasattr(self, 'webfrontend_version'):
+            yield self.webfrontend, self.webfrontend_version
 
     def component_activated(self, component):
         """Initialize additional member variables for components.
@@ -338,9 +365,8 @@ class Environment(Component, ComponentManager):
     def _component_rules(self):
         _rules = {}
         for name, value in self.components_section.options():
-            if name.endswith('.*'):
-                name = name[:-2]
-            _rules[name.lower()] = value.lower() in ('enabled', 'on')
+            name = name.rstrip('.*').lower()
+            _rules[name] = as_bool(value)
         return _rules
 
     def is_component_enabled(self, cls):
@@ -355,18 +381,6 @@ class Environment(Component, ComponentManager):
         """
         component_name = self._component_name(cls)
 
-        # Disable the pre-0.11 WebAdmin plugin
-        # Please note that there's no recommendation to uninstall the
-        # plugin because doing so would obviously break the backwards
-        # compatibility that the new integration administration
-        # interface tries to provide for old WebAdmin extensions
-        if component_name.startswith('webadmin.'):
-            self.log.info("The legacy TracWebAdmin plugin has been "
-                          "automatically disabled, and the integrated "
-                          "administration interface will be used "
-                          "instead.")
-            return False
-
         rules = self._component_rules
         cname = component_name
         while cname:
@@ -379,13 +393,15 @@ class Environment(Component, ComponentManager):
             cname = cname[:idx]
 
         # By default, all components in the trac package except
-        # trac.test are enabled
+        # in trac.test or trac.tests are enabled
         return component_name.startswith('trac.') and \
-               not component_name.startswith('trac.test.') or None
+               not component_name.startswith('trac.test.') and \
+               not component_name.startswith('trac.tests.') or None
 
     def enable_component(self, cls):
         """Enable a component or module."""
         self._component_rules[self._component_name(cls)] = True
+        super(Environment, self).enable_component(cls)
 
     def verify(self):
         """Verify that the provided path points to a valid Trac environment
@@ -395,37 +411,10 @@ class Environment(Component, ComponentManager):
             if tag != _VERSION:
                 raise Exception(_("Unknown Trac environment type '%(type)s'",
                                   type=tag))
-        except Exception, e:
+        except Exception as e:
             raise TracError(_("No Trac environment found at %(path)s\n"
                               "%(e)s", path=self.path, e=e))
 
-    def get_db_cnx(self):
-        """Return a database connection from the connection pool
-
-        :deprecated: Use :meth:`db_transaction` or :meth:`db_query` instead.
-                     Removed in Trac 1.1.2.
-
-        `db_transaction` for obtaining the `db` database connection
-        which can be used for performing any query
-        (SELECT/INSERT/UPDATE/DELETE)::
-
-           with env.db_transaction as db:
-               ...
-
-        Note that within the block, you don't need to (and shouldn't)
-        call ``commit()`` yourself, the context manager will take care
-        of it (if it's the outermost such context manager on the
-        stack).
-
-
-        `db_query` for obtaining a `db` database connection which can
-        be used for performing SELECT queries only::
-
-           with env.db_query as db:
-               ...
-        """
-        return DatabaseManager(self).get_connection()
-
     @lazy
     def db_exc(self):
         """Return an object (typically a module) containing all the
@@ -438,7 +427,7 @@ class Environment(Component, ComponentManager):
             try:
                 with env.db_transaction as db:
                     ...
-            except env.db_exc.IntegrityError, e:
+            except env.db_exc.IntegrityError as e:
                 ...
         """
         return DatabaseManager(self).get_exceptions()
@@ -539,15 +528,16 @@ class Environment(Component, ComponentManager):
 
     def shutdown(self, tid=None):
         """Close the environment."""
+        from trac.versioncontrol.api import RepositoryManager
         RepositoryManager(self).shutdown(tid)
         DatabaseManager(self).shutdown(tid)
-        if tid is None:
+        if tid is None and self._log_handler is not None:
             self.log.removeHandler(self._log_handler)
             self._log_handler.flush()
             self._log_handler.close()
             del self._log_handler
 
-    def get_repository(self, reponame=None, authname=None):
+    def get_repository(self, reponame=None):
         """Return the version control repository with the given name,
         or the default repository if `None`.
 
@@ -556,10 +546,10 @@ class Environment(Component, ComponentManager):
         for backward compatibility.
 
         :param reponame: the name of the repository
-        :param authname: the user name for authorization (not used
-                         anymore, left here for compatibility with
-                         0.11)
+
+        :since 1.2: deprecated and will be removed in 1.3.1
         """
+        from trac.versioncontrol.api import RepositoryManager
         return RepositoryManager(self).get_repository(reponame)
 
     def create(self, options=[]):
@@ -574,9 +564,9 @@ class Environment(Component, ComponentManager):
         # Create the directory structure
         if not os.path.exists(self.path):
             os.mkdir(self.path)
-        os.mkdir(self.get_log_dir())
-        os.mkdir(self.get_htdocs_dir())
-        os.mkdir(os.path.join(self.path, 'plugins'))
+        os.mkdir(self.log_dir)
+        os.mkdir(self.htdocs_dir)
+        os.mkdir(self.plugins_dir)
 
         # Create a few files
         create_file(os.path.join(self.path, 'VERSION'), _VERSION + '\n')
@@ -585,9 +575,9 @@ class Environment(Component, ComponentManager):
                     'Visit http://trac.edgewall.org/ for more information.\n')
 
         # Setup the default configuration
-        os.mkdir(os.path.join(self.path, 'conf'))
-        create_file(os.path.join(self.path, 'conf', 'trac.ini.sample'))
-        config = Configuration(os.path.join(self.path, 'conf', 'trac.ini'))
+        os.mkdir(self.conf_dir)
+        create_file(self.config_file_path + '.sample')
+        config = Configuration(self.config_file_path)
         for section, name, value in options:
             config.set(section, name, value)
         config.save()
@@ -606,7 +596,8 @@ class Environment(Component, ComponentManager):
 
         :since 1.0.2:
         """
-        return self.get_version()
+        return DatabaseManager(self) \
+               .get_database_version('database_version')
 
     @lazy
     def database_initial_version(self):
@@ -617,9 +608,18 @@ class Environment(Component, ComponentManager):
 
         :since 1.0.2:
         """
-        return self.get_version(initial=True)
+        return DatabaseManager(self) \
+               .get_database_version('initial_database_version')
 
-    def get_version(self, db=None, initial=False):
+    @lazy
+    def trac_version(self):
+        """Returns the version of Trac.
+        :since: 1.2
+        """
+        from trac import core, __version__
+        return get_pkginfo(core).get('version', __version__)
+
+    def get_version(self, initial=False):
         """Return the current version of the database.  If the
         optional argument `initial` is set to `True`, the version of
         the database used at the time of creation will be returned.
@@ -629,75 +629,130 @@ class Environment(Component, ComponentManager):
 
         :since: 0.11
 
-        :since 1.0: deprecation warning: the `db` parameter is no
-                    longer used and will be removed in version 1.1.1
-
         :since 1.0.2: The lazily-evaluated attributes `database_version` and
                       `database_initial_version` should be used instead. This
-                      method will be renamed to a private method in
-                      release 1.3.1.
+                      method will be removed in release 1.3.1.
         """
-        rows = self.db_query("""
-                SELECT value FROM system WHERE name='%sdatabase_version'
-                """ % ('initial_' if initial else ''))
-        return int(rows[0][0]) if rows else False
+        dbm = DatabaseManager(self)
+        return dbm.get_database_version(
+            '{0}database_version'.format('initial_' if initial else ''))
 
     def setup_config(self):
         """Load the configuration file."""
-        self.config = Configuration(os.path.join(self.path, 'conf', 'trac.ini'),
-                                    {'envname': os.path.basename(self.path)})
+        self.config = Configuration(self.config_file_path,
+                                    {'envname': self.name})
+        if not self.config.exists:
+            raise TracError(_("The configuration file is not found at "
+                              "%(path)s", path=self.config_file_path))
         self.setup_log()
-        from trac.loader import load_components
         plugins_dir = self.shared_plugins_dir
         load_components(self, plugins_dir and (plugins_dir,))
 
-    def get_templates_dir(self):
-        """Return absolute path to the templates directory."""
-        return os.path.join(self.path, 'templates')
+    @lazy
+    def config_file_path(self):
+        """Path of the trac.ini file."""
+        return os.path.join(self.conf_dir, 'trac.ini')
+
+    def _get_path_to_dir(self, dir):
+        path = os.path.join(self.path, dir)
+        return os.path.normcase(os.path.realpath(path))
+
+    @lazy
+    def conf_dir(self):
+        """Absolute path to the conf directory.
+
+        :since: 1.0.11
+        """
+        return self._get_path_to_dir('conf')
+
+    @lazy
+    def htdocs_dir(self):
+        """Absolute path to the htdocs directory.
+
+        :since: 1.0.11
+        """
+        return self._get_path_to_dir('htdocs')
 
     def get_htdocs_dir(self):
-        """Return absolute path to the htdocs directory."""
-        return os.path.join(self.path, 'htdocs')
+        """Return absolute path to the htdocs directory.
+
+        :since 1.0.11: Deprecated and will be removed in 1.3.1. Use the
+                       `htdocs_dir` property instead.
+        """
+        return self._get_path_to_dir('htdocs')
+
+    @lazy
+    def log_dir(self):
+        """Absolute path to the log directory.
+
+        :since: 1.0.11
+        """
+        return self._get_path_to_dir('log')
 
     def get_log_dir(self):
-        """Return absolute path to the log directory."""
-        return os.path.join(self.path, 'log')
+        """Return absolute path to the log directory.
+
+        :since 1.0.11: Deprecated and will be removed in 1.3.1. Use the
+                       `log_dir` property instead.
+        """
+        return self._get_path_to_dir('log')
+
+    @lazy
+    def plugins_dir(self):
+        """Absolute path to the plugins directory.
+
+        :since: 1.0.11
+        """
+        return self._get_path_to_dir('plugins')
+
+    @lazy
+    def templates_dir(self):
+        """Absolute path to the templates directory.
+
+        :since: 1.0.11
+        """
+        return self._get_path_to_dir('templates')
+
+    def get_templates_dir(self):
+        """Return absolute path to the templates directory.
+
+        :since 1.0.11: Deprecated and will be removed in 1.3.1. Use the
+                       `templates_dir` property instead.
+        """
+        return self._get_path_to_dir('templates')
 
     def setup_log(self):
         """Initialize the logging sub-system."""
-        from trac.log import logger_handler_factory
         logtype = self.log_type
         logfile = self.log_file
         if logtype == 'file' and not os.path.isabs(logfile):
-            logfile = os.path.join(self.get_log_dir(), logfile)
+            logfile = os.path.join(self.log_dir, logfile)
         format = self.log_format
-        logid = 'Trac.%s' % sha1(self.path).hexdigest()
+        logid = 'Trac.%s' % hashlib.sha1(self.path).hexdigest()
         if format:
             format = format.replace('$(', '%(') \
-                     .replace('%(path)s', self.path) \
-                     .replace('%(basename)s', os.path.basename(self.path)) \
-                     .replace('%(project)s', self.project_name)
+                           .replace('%(path)s', self.path) \
+                           .replace('%(basename)s', self.name) \
+                           .replace('%(project)s', self.project_name)
         self.log, self._log_handler = logger_handler_factory(
             logtype, logfile, self.log_level, logid, format=format)
-        from trac import core, __version__ as VERSION
         self.log.info('-' * 32 + ' environment startup [Trac %s] ' + '-' * 32,
-                      get_pkginfo(core).get('version', VERSION))
+                      self.trac_version)
 
-    def get_known_users(self, cnx=None):
-        """Generator that yields information about all known users,
-        i.e. users that have logged in to this Trac environment and
-        possibly set their name and email.
+    def get_known_users(self, as_dict=False):
+        """Returns information about all known users, i.e. users that
+        have logged in to this Trac environment and possibly set their
+        name and email.
 
-        This function generates one tuple for every user, of the form
-        (username, name, email) ordered alpha-numerically by username.
+        By default this function returns a iterator that yields one
+        tuple for every user, of the form (username, name, email),
+        ordered alpha-numerically by username. When `as_dict` is `True`
+        the function returns a dictionary mapping username to a
+        (name, email) tuple.
 
-        :param cnx: the database connection; if ommitted, a new
-                    connection is retrieved
-
-        :since 1.0: deprecation warning: the `cnx` parameter is no
-                    longer used and will be removed in version 1.1.1
+        :since 1.2: the `as_dict` parameter is available.
         """
-        return iter(self._known_users)
+        return self._known_users_dict if as_dict else iter(self._known_users)
 
     @cached
     def _known_users(self):
@@ -711,9 +766,14 @@ class Environment(Component, ComponentManager):
                 WHERE s.authenticated=1 ORDER BY s.sid
         """)
 
+    @cached
+    def _known_users_dict(self):
+        return dict([(u[0], (u[1], u[2])) for u in self._known_users])
+
     def invalidate_known_users_cache(self):
         """Clear the known_users cache."""
         del self._known_users
+        del self._known_users_dict
 
     def backup(self, dest=None):
         """Create a backup of the database.
@@ -726,8 +786,11 @@ class Environment(Component, ComponentManager):
     def needs_upgrade(self):
         """Return whether the environment needs to be upgraded."""
         for participant in self.setup_participants:
+            args = ()
             with self.db_query as db:
-                if participant.environment_needs_upgrade(db):
+                if arity(participant.environment_needs_upgrade) == 1:
+                    args = (db,)
+                if participant.environment_needs_upgrade(*args):
                     self.log.warn("Component %s requires environment upgrade",
                                   participant)
                     return True
@@ -742,8 +805,11 @@ class Environment(Component, ComponentManager):
         """
         upgraders = []
         for participant in self.setup_participants:
+            args = ()
             with self.db_query as db:
-                if participant.environment_needs_upgrade(db):
+                if arity(participant.environment_needs_upgrade) == 1:
+                    args = (db,)
+                if participant.environment_needs_upgrade(*args):
                     upgraders.append(participant)
         if not upgraders:
             return
@@ -751,14 +817,17 @@ class Environment(Component, ComponentManager):
         if backup:
             try:
                 self.backup(backup_dest)
-            except Exception, e:
+            except Exception as e:
                 raise BackupError(e)
 
         for participant in upgraders:
             self.log.info("%s.%s upgrading...", participant.__module__,
                           participant.__class__.__name__)
+            args = ()
             with self.db_transaction as db:
-                participant.upgrade_environment(db)
+                if arity(participant.upgrade_environment) == 1:
+                    args = (db,)
+                participant.upgrade_environment(*args)
             # Database schema may have changed, so close all connections
             DatabaseManager(self).shutdown()
         del self.database_version
@@ -789,49 +858,21 @@ class EnvironmentSetup(Component):
 
     def environment_created(self):
         """Insert default data into the database."""
-        with self.env.db_transaction as db:
-            for table, cols, vals in db_default.get_data(db):
-                db.executemany("INSERT INTO %s (%s) VALUES (%s)"
-                   % (table, ','.join(cols), ','.join(['%s' for c in cols])),
-                   vals)
+        DatabaseManager(self.env).insert_into_tables(db_default.get_data)
         self._update_sample_config()
 
-    def environment_needs_upgrade(self, db):
-        dbver = self.env.database_version
-        if dbver == db_default.db_version:
-            return False
-        elif dbver > db_default.db_version:
-            raise TracError(_('Database newer than Trac version'))
-        self.log.info("Trac database schema version is %d, should be %d",
-                      dbver, db_default.db_version)
-        return True
+    def environment_needs_upgrade(self):
+        return DatabaseManager(self.env).needs_upgrade(db_default.db_version)
 
-    def upgrade_environment(self, db):
-        """Each db version should have its own upgrade module, named
-        upgrades/dbN.py, where 'N' is the version number (int).
-        """
-        cursor = db.cursor()
-        dbver = self.env.database_version
-        for i in range(dbver + 1, db_default.db_version + 1):
-            name  = 'db%i' % i
-            try:
-                upgrades = __import__('upgrades', globals(), locals(), [name])
-                script = getattr(upgrades, name)
-            except AttributeError:
-                raise TracError(_("No upgrade module for version %(num)i "
-                                  "(%(version)s.py)", num=i, version=name))
-            script.do_upgrade(self.env, i, cursor)
-            cursor.execute("""
-                UPDATE system SET value=%s WHERE name='database_version'
-                """, (i,))
-            self.log.info("Upgraded database version from %d to %d", i - 1, i)
-            db.commit()
+    def upgrade_environment(self):
+        DatabaseManager(self.env).upgrade(db_default.db_version,
+                                          pkg='trac.upgrades')
         self._update_sample_config()
 
     # Internal methods
 
     def _update_sample_config(self):
-        filename = os.path.join(self.env.path, 'conf', 'trac.ini.sample')
+        filename = os.path.join(self.env.config_file_path + '.sample')
         if not os.path.isfile(filename):
             return
         config = Configuration(filename)
@@ -842,7 +883,7 @@ class EnvironmentSetup(Component):
             self.log.info("Wrote sample configuration file with the new "
                           "settings and their default values: %s",
                           filename)
-        except IOError, e:
+        except IOError as e:
             self.log.warn("Couldn't write sample configuration file (%s)", e,
                           exc_info=True)
 
@@ -850,12 +891,13 @@ class EnvironmentSetup(Component):
 env_cache = {}
 env_cache_lock = threading.Lock()
 
+
 def open_environment(env_path=None, use_cache=False):
     """Open an existing environment object, and verify that the database is up
     to date.
 
     :param env_path: absolute path to the environment directory; if
-                     ommitted, the value of the `TRAC_ENV` environment
+                     omitted, the value of the `TRAC_ENV` environment
                      variable is used
     :param use_cache: whether the environment should be cached for
                       subsequent invocations of this function
@@ -881,7 +923,8 @@ def open_environment(env_path=None, use_cache=False):
                 del env_cache[env_path]
                 env = None
             if env is None:
-                env = env_cache.setdefault(env_path, open_environment(env_path))
+                env = env_cache.setdefault(env_path,
+                                           open_environment(env_path))
             else:
                 CacheManager(env).reset_metadata()
     else:
@@ -889,7 +932,7 @@ def open_environment(env_path=None, use_cache=False):
         needs_upgrade = False
         try:
             needs_upgrade = env.needs_upgrade()
-        except Exception, e: # e.g. no database connection
+        except Exception as e:  # e.g. no database connection
             env.log.error("Exception caught while checking for upgrade: %s",
                           exception_to_unicode(e, traceback=True))
         if needs_upgrade:
@@ -919,8 +962,13 @@ class EnvironmentAdmin(Component):
                specified.
                """,
                None, self._do_hotcopy)
-        yield ('upgrade', '',
-               'Upgrade database to current version',
+        yield ('upgrade', '[--no-backup]',
+               """Upgrade database to current version
+
+               The database is backed up to the environment 'db' directory,
+               unless the --no-backup option is specified. The shorthand alias
+               -b can also be used to specify --no-backup.
+               """,
                None, self._do_upgrade)
 
     def _do_deploy(self, dest):
@@ -968,7 +1016,6 @@ class EnvironmentAdmin(Component):
         if os.path.exists(dest):
             raise TracError(_("hotcopy can't overwrite existing '%(dest)s'",
                               dest=path_to_unicode(dest)))
-        import shutil
 
         # Bogus statement to lock the database while copying files
         with self.env.db_transaction as db:
@@ -982,16 +1029,18 @@ class EnvironmentAdmin(Component):
             skip = []
 
             if prefix == 'sqlite':
-                db_path = os.path.join(self.env.path, os.path.normpath(db_path))
+                db_path = os.path.join(self.env.path,
+                                       os.path.normpath(db_path))
                 # don't copy the journal (also, this would fail on Windows)
-                skip = [db_path + '-journal', db_path + '-stmtjrnl']
+                skip = [db_path + '-journal', db_path + '-stmtjrnl',
+                        db_path + '-shm', db_path + '-wal']
                 if no_db:
                     skip.append(db_path)
 
             try:
                 copytree(self.env.path, dest, symlinks=1, skip=skip)
                 retval = 0
-            except shutil.Error, e:
+            except shutil.Error as e:
                 retval = 1
                 printerr(_("The following errors happened while copying "
                            "the environment:"))
@@ -1001,7 +1050,6 @@ class EnvironmentAdmin(Component):
                     else:
                         printerr("  %s: '%s'" % (err, path_to_unicode(src)))
 
-
             # db backup for non-sqlite
             if prefix != 'sqlite' and not no_db:
                 printout(_("Backing up database ..."))
@@ -1022,11 +1070,11 @@ class EnvironmentAdmin(Component):
 
         try:
             self.env.upgrade(backup=no_backup is None)
-        except BackupError, e:
+        except BackupError as e:
             printerr(_("The pre-upgrade backup failed.\nUse '--no-backup' to "
                        "upgrade without doing a backup.\n"))
             raise e.args[0]
-        except Exception, e:
+        except Exception:
             printerr(_("The upgrade failed. Please fix the issue and try "
                        "again.\n"))
             raise
@@ -1046,7 +1094,7 @@ class EnvironmentAdmin(Component):
             else:
                 try:
                     os.rmdir(wiki_macros)
-                except OSError, e:
+                except OSError as e:
                     printerr(_("Error while removing wiki-macros: %(err)s\n"
                                "Trac doesn't load plugins from wiki-macros "
                                "anymore. Please remove it by hand.",
